__all__ = [
    "BodyMetadataUnpack",
    "ContentExample",
    "RequestExample",
    "ResponseExample",
    "RouteData",
    "RouteMetadataUnpack",
]
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Self, TypedDict

from fastapi.openapi.models import Example
from pydantic import AnyUrl, BaseModel

from kyotsu.utils import JsonAlias


class BodyMetadataUnpack(TypedDict, total=False):
    title: str | None
    description: str | None
    openapi_examples: dict[str, Example] | None


class RouteMetadataUnpack(TypedDict, total=False):
    name: str | None
    summary: str | None
    description: str | None
    response_description: str
    tags: list[str | Enum] | None
    deprecated: bool | None
    responses: dict[str | int, dict[str, Any]] | None


@dataclass
class RequestExample:
    name: str
    summary: str
    description: str | None = None
    value: JsonAlias = None
    external_value: AnyUrl | None = None


@dataclass
class ContentExample:
    name: str
    value: JsonAlias = None


@dataclass
class ResponseExample:
    status_code: str
    description: str
    model: type[BaseModel] | None = None
    examples: list[ContentExample] | None = None


@dataclass(kw_only=True)
class RouteData:
    route_name: str | None = None
    route_summary: str | None = None
    route_description: str | None = None
    route_response_description: str = "Successful Response"
    route_tags: list[str | Enum] | None = None
    route_deprecated: bool = False
    _route_responses: list[ResponseExample] | None = None
    route_responses: dict[str | int, dict[str, JsonAlias | type[BaseModel]]] | None = field(init=False)

    body_title: str | None = None
    body_description: str | None = None
    _body_examples: list[RequestExample] | None = None
    body_openapi_examples: dict[str, Example] | None = field(init=False)

    @staticmethod
    def _get_response_example(example: ResponseExample) -> dict[str, JsonAlias | type[BaseModel]]:
        result: dict[str, JsonAlias | type[BaseModel]] = {"description": example.description}
        if example.model:
            result["model"] = example.model
        if example.examples:
            result["content"] = {
                "application/json": {
                    "examples": {content.name: {"value": content.value} for content in example.examples},
                },
            }
        return result

    def __post_init__(self: Self) -> None:
        if self._route_responses is None:
            self.route_responses = None
        else:
            self.route_responses = {
                response.status_code: self._get_response_example(response) for response in self._route_responses
            }
        if self._body_examples is None:
            self.body_openapi_examples = None
        else:
            self.body_openapi_examples = {
                example.name: Example(
                    summary=example.summary,
                    description=example.description,
                    value=example.value,
                    externalValue=example.external_value,
                )
                for example in self._body_examples
            }

    @property
    def route_metadata(self: Self) -> RouteMetadataUnpack:
        return RouteMetadataUnpack(
            name=self.route_name,
            summary=self.route_summary,
            description=self.route_description,
            response_description=self.route_response_description,
            tags=self.route_tags,
            deprecated=self.route_deprecated,
            responses=self.route_responses,
        )

    @property
    def body_metadata(self: Self) -> BodyMetadataUnpack:
        return BodyMetadataUnpack(
            title=self.body_title,
            description=self.body_description,
            openapi_examples=self.body_openapi_examples,
        )



"""This module contains keeper class for uncategorized errors."""

__all__ = [
    "UncategorizedKeeper",
]

from typing import Final

from kyotsu.logging.error_codes.dataclasses import ErrorCode, error_code_keeper

from ._constants import Prefix


@error_code_keeper
class UncategorizedKeeper:
    """Collection of uncategorized errors.

    Danger:
        Do not add anything to this category unless you are absolutely sure in your actions.
        # ToDo: Add link to recipes
        See: [Using Error Prefix]()

    Attributes:
        UNHANDLED_EXCEPTION: Error code that indicates previously unhandled exception.
    """

    UNHANDLED_EXCEPTION: Final[ErrorCode] = ErrorCode(
        Prefix.UN,
        1
        # _dev_description=(
        #     "Unhandled error has occurred. Please consider validating an error source and add error handler."
        # ),
        # _hp_description="Something bad happened. Please contact developers for further investigation.",
    )

    REQUEST_VALIDATION_ERROR: Final[ErrorCode] = ErrorCode(
        Prefix.UN,
        2,
        message="Unprocessable entity in request",
        # _dev_description=(
        #     "FastAPI request validation error has occurred. "
        #     "Consider validating request/query params/etc. over pydantic models."
        # ),
        # _hp_description=(
        #     "At some point of Frontend-to-Service or Service-to-Service communication "
        #     "some part of request/query params/etc. appeared inconsistent with API contract. "
        #     "This can be caused by incorrect user input or by poor API contract handling. "
        #     "Send this log to dev team for further investigation."
        # ),
    )

    RESPONSE_VALIDATION_ERROR: Final[ErrorCode] = ErrorCode(
        Prefix.UN,
        3,
        # _dev_description=(
        #     "FastAPI response validation error has occurred. Consider validating response. over pydantic models."
        # ),
        # _hp_description=(
        #     "At some point backend returned invalid response (inconsistent with API contract). "
        #     "Send this log to dev team for further investigation."
        # ),
    )



__all__ = [
    "SyncAsyncContextDecorator",
]
import abc
from asyncio import iscoroutinefunction
from collections.abc import Awaitable, Callable
from functools import wraps
from types import TracebackType
from typing import Any, Self


class SyncAsyncContextDecorator(abc.ABC):
    def _recreate_cm(self: Self) -> Self:
        """
        Return a recreated instance of self.

        Copied from [contextlib.ContextDecorator][].
        """
        return self

    def __call__(
        self,
        func: Callable[..., Any] | Callable[..., Awaitable[Any]],
    ) -> Callable[..., Any] | Callable[..., Awaitable[Any]]:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            with self._recreate_cm():
                return func(*args, **kwargs)

        @wraps(func)
        async def awrapper(*args: Any, **kwargs: Any) -> Any:
            async with self._recreate_cm():
                return await func(*args, **kwargs)

        if not iscoroutinefunction(func):
            return wrapper
        return awrapper

    @abc.abstractmethod
    def __enter__(self: Self) -> Self: ...

    @abc.abstractmethod
    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool: ...

    async def __aenter__(self: Self) -> Self:
        return self.__enter__()

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool:
        return self.__exit__(exc_type, exc_val, exc_tb)



__all__ = [
    "ErrorCodesKeeper",
]

from typing import Final

from kyotsu.logging.error_codes.dataclasses import error_code_keeper

from .kyotsu import KyotsuKeeper
from .scheduler_service import SchedulerKeeper
from .uncategorized import UncategorizedKeeper


@error_code_keeper
class ErrorCodesKeeper:
    """
    A single keeper for all error codes.
    Instance of this class serves as an entrypoint to the :sparkles:error codes world:sparkles:.


    To introduce a new category simply add new class attribute with
    Prefix as attribute name, `Final[<your_keeper_class>]` as attribute type,
    and your keeper class instance as attribute value.

    Attributes:
        UN: Keeps uncategorized error codes.
        KTS: Keeps kyotsu runtime error codes.
    """

    UN: Final[UncategorizedKeeper] = UncategorizedKeeper()
    KTS: Final[KyotsuKeeper] = KyotsuKeeper()
    SCH: Final[SchedulerKeeper] = SchedulerKeeper()



__all__ = [
    "error_code_adder",
]

from typing import TYPE_CHECKING

from structlog.processors import _figure_out_exc_info
from structlog.typing import EventDict, WrappedLogger

if TYPE_CHECKING:
    from kyotsu.logging.error_codes import ErrorCode


def error_code_adder(_: WrappedLogger, __: str, event_dict: EventDict) -> EventDict:
    exc_info = _figure_out_exc_info(event_dict.get("exc_info"))

    if not exc_info or all(e is None for e in exc_info):
        return event_dict

    try:
        code: "ErrorCode" = exc_info[1].error_code  # type: ignore[attr-defined] # We are handling AttributeError
    except AttributeError:
        return event_dict

    error_details = {
        "code": code.code,
        "message": code.message,
    }

    try:
        event_dict["error_details"] |= error_details
    except KeyError:
        event_dict["error_details"] = error_details

    return event_dict



__all__ = [
    "AsyncSessionMaker",
    "SessionMaker",
]

import abc
from typing import Any, Generic, Self, TypeVar

from sqlalchemy import Engine, create_engine
from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession, async_sessionmaker, create_async_engine
from sqlalchemy.orm import Session, sessionmaker

from kyotsu.config.databases import PostgresConnection

TEngine = TypeVar("TEngine", Engine, AsyncEngine)
TSession = TypeVar("TSession", Session, AsyncSession)


class AbstractSessionMaker(abc.ABC, Generic[TEngine, TSession]):
    ENGINE: str

    def __init__(
        self: Self,
        conn_settings: PostgresConnection,
        *,
        rebind_new_engine: bool = False,
        new_engine_kwargs: dict[str, Any] | None = None,
        **session_maker_kwargs: Any,
    ):
        self._new_engine_kwargs = new_engine_kwargs or {}
        self._conn_str = conn_settings.CONN_STR.replace("postgresql", f"postgresql+{self.ENGINE}")
        self._rebind_new_engine = rebind_new_engine
        self._session_maker_kwargs = session_maker_kwargs
        self._engine: TEngine | None = self.get_new_engine() if not self._rebind_new_engine else None

    @abc.abstractmethod
    def __call__(self: Self) -> TSession: ...

    @abc.abstractmethod
    def get_new_engine(self: Self) -> TEngine: ...


class SessionMaker(AbstractSessionMaker[Engine, Session]):
    ENGINE = "psycopg2"

    def __call__(self: Self) -> Session:
        engine = self._engine if self._engine is not None else self.get_new_engine()
        session: Session = sessionmaker(bind=engine, **self._session_maker_kwargs)()
        return session

    def get_new_engine(self: Self) -> Engine:
        engine: Engine = create_engine(self._conn_str, **self._new_engine_kwargs)
        return engine


class AsyncSessionMaker(AbstractSessionMaker[AsyncEngine, AsyncSession]):
    ENGINE = "asyncpg"

    def __call__(self: Self) -> AsyncSession:
        engine = self._engine if self._engine is not None else self.get_new_engine()
        session: AsyncSession = async_sessionmaker(bind=engine, **self._session_maker_kwargs)()
        return session

    def get_new_engine(self: Self) -> AsyncEngine:
        engine: AsyncEngine = create_async_engine(self._conn_str, **self._new_engine_kwargs)
        return engine



__all__ = [
    "parent_span_id",
    "span_id",
    "trace_id",
]

from contextvars import ContextVar

trace_id: ContextVar[str | None] = ContextVar("trace_id", default=None)
span_id: ContextVar[str | None] = ContextVar("span_id", default=None)
parent_span_id: ContextVar[str | None] = ContextVar("parent_span_id", default=None)



"""
This module contains base models for defining database connection settings.

Each model class in this module represents a different type of database connection.

All classes derived from `DBConnection` that use separate properties
(`USERNAME`, `PASSWORD`, `HOST`, `PORT`, `DATABASE`) to define a connection string (`CONN_STR`) use two main methods:

- `_assemble_conn_str_value()`: an abstract method, which needs to be defined in subclasses.
It is responsible for creating the actual connection string value.

- `assemble_conn_str()`: a model validator, which checks whether `CONN_STR` is `None`.
If yes, it builds the connection string using `_assemble_conn_str_value()`.

Classes:
    DBConnection: Abstract class that can be inherited to represent different types of DB connections.
    AmqpConnection: Represents an AMQP (Advanced Message Queuing Protocol) connection.
    OpenSearchNoCertConnection: Represents an OpenSearch connection without requiring SSL certification.
    PostgresConnection: Represents a connection to a Postgres database using either asyncpg or psycopg2 scheme.
    RedisConnection: Represents a Redis database connection string with support for SSL.
"""

__all__ = [
    "AmqpConnection",
    "DBConnection",
    "OpenSearchNoCertConnection",
    "PostgresConnection",
    "RedisConnection",
]

import abc
import warnings
import urllib.parse
from typing import Annotated, Generic, Literal, Self

from pydantic import AmqpDsn, AnyHttpUrl, BaseModel, ConfigDict, PostgresDsn, RedisDsn, computed_field

from kyotsu.config.databases._types import AmqpQueryParams, ConnType, OpenSearchConnectionDict


class DBConnection(BaseModel, Generic[ConnType], abc.ABC):
    """
    This abstract base model class represents a generic database connection.

    It is designed to be inherited by subclasses that represent specific database connections.

    Subclasses must implement the `_assemble_conn_str_value` method to generate the necessary connection string
    representation for a specific database type.

    Attributes:
        USERNAME (str | None): The username for the database connection.
        PASSWORD (str | None): The password for the database connection.
        HOST (str | None): The host for the database connection.
        PORT (str | None): The port for the database connection.
        DATABASE (str | None): The specific database to connect to.
        CONN_STR (Annotated[str, ConnType]): The connection string, assembled from the other properties.
    """

    USERNAME: str | None = None
    PASSWORD: str | None = None
    HOST: str | None = None
    PORT: int | None = None
    DATABASE: str | None = None

    @computed_field
    @property
    def CONN_STR(self: Self) -> Annotated[str, ConnType]:
        """
        The connection string, assembled from the other properties.

        This attribute is computed by calling the `_assemble_conn_str_value` method.

        Returns:
            The connection string, assembled from the other properties.
        """
        return str(self._assemble_conn_str_value())

    @abc.abstractmethod
    def _assemble_conn_str_value(self: Self) -> ConnType:
        """
        Abstract method that subclasses must implement to generate a connection string representation.

        Arguments:
            self (Self): A reference to the instance of a subclass that this method will be called on.

        Returns:
            ConnType: A connection string that can be directly used by the specific type of database client
                to create a connection with the database.
        """
        ...


class AmqpConnection(DBConnection[AmqpDsn]):
    """
    Represents an AMQP (Advanced Message Queuing Protocol) connection.

    Attributes:
        USERNAME (str): The username for this AMQP service.
        PASSWORD (str): The password for this AMQP service.
        HOST (str): The host of the AMQP service.
        PORT (int): [Optional] The port of the AMQP service. Default is 5672.
        DATABASE (str): [Optional] The database for this AMQP service. Default is "%2f", a placeholder for root level.
        USE_SSL (bool): [Optional] If true, the AMQP service is assumed to use SSL. Default is True.
        QUERY_PARAMS (AmqpQueryParams | None): [Optional] Set of optional parameters for the connection,
            including backpressure detection, channel max, etc.

        CONN_STR (str): The connection string for the AMQP service, assembled from the other properties.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    USERNAME: str
    PASSWORD: str
    HOST: str
    PORT: int = 5672
    # "%2f" value is used as default virtual host
    # https://pika.readthedocs.io/en/stable/modules/parameters.html#pika.connection.URLParameters
    DATABASE: str = "%2f"

    USE_SSL: bool = True
    QUERY_PARAMS: AmqpQueryParams | None = None

    def _assemble_conn_str_value(self: Self) -> AmqpDsn:
        """
        Assembles a connection string representation for an AMQP connection.

        Arguments:
            self (Self): A reference to the instance of AmqpConnection that this method will be called on.

        Returns:
            AmqpDsn: A Pydantic AMQP DSN model that represents the connection string.
        """
        scheme = "amqps" if self.USE_SSL else "amqp"
        query_params = urllib.parse.urlencode(self.QUERY_PARAMS) if self.QUERY_PARAMS else None
        return AmqpDsn.build(
            scheme=scheme,
            password=self.PASSWORD,
            host=self.HOST.split("://")[-1],
            port=self.PORT,
            path=self.DATABASE,
            query=query_params,
        )


class OpenSearchNoCertConnection(DBConnection[AnyHttpUrl]):
    """
    Represents an OpenSearch connection without requiring SSL certification.

    Attributes:
        SCHEME (Literal["http", "https"]): The scheme for this connection, either "http" or "https".
        USERNAME (str): The username for the OpenSearch service.
        PASSWORD (str): The password for the OpenSearch service.
        HOST (str): The host of the OpenSearch service.
        PORT (int): The port of the OpenSearch service.
        TTL (int): [Optional] The duration after which the connection times out. Default is 60.
        MAX_RETRIES (int): [Optional] The maximum number of times the request to the connection can be retried.
            Default is 10.
        RETRY_ON_TIMEOUT (bool): [Optional] If true, retries occur on timed-out connections. Default is True.
        CONN_DICT (OpenSearchConnectionDict): After being processed in `assemble_conn_dict`, this dict contains
                                                details about host, authentication, connectivity, and SSL context.

        CONN_STR (str): The connection string for the OpenSearch service, assembled from,
                                        assembled from the other properties.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    SCHEME: Literal["http", "https"]
    USERNAME: str
    PASSWORD: str
    HOST: str
    PORT: int

    TTL: int = 60
    MAX_RETRIES: int = 10
    RETRY_ON_TIMEOUT: bool = True

    @computed_field
    @property
    def CONN_DICT(self: Self) -> OpenSearchConnectionDict:
        """
        Returns dict with OpenSearch connection parameters,
        including host, scheme, port, authentication, and SSL context.

        Warns:
            ImportWarning: When opensearch-py is not installed, therefore the ssl_context can't be added.

        Returns:
            OpenSearchConnectionDict: The connection dictionary.
        """
        conn_dict = OpenSearchConnectionDict(
            hosts=(self.HOST,),
            http_auth=(self.USERNAME, self.PASSWORD) if self.USERNAME else None,
            scheme=self.SCHEME,
            port=str(self.PORT),
            timeout=self.TTL,
            max_retries=self.MAX_RETRIES,
            retry_on_timeout=self.RETRY_ON_TIMEOUT,
        )

        import ssl

        try:
            from opensearchpy.connection import create_ssl_context
        except ImportError:
            msg = "Can't add SSL context to OpenSearch Connection Dict. Make sure that opensearch-py is installed."
            warnings.warn(msg, ImportWarning, stacklevel=1)
            return conn_dict

        ssl_context = create_ssl_context()
        ssl_context.check_hostname = False
        ssl_context.verify_mode = ssl.CERT_NONE

        conn_dict["ssl_context"] = ssl_context
        return conn_dict

    def _assemble_conn_str_value(self: Self) -> AnyHttpUrl:
        """
        Assembles a connection string representation for an OpenSearch connection.

        Arguments:
            self (Self): A reference to the instance of `OpenSearchNoCertConnection` that this method will be called on.

        Returns:
            AnyHttpUrl: A Pydantic Http URL model (`AnyHttpUrl`) that represents the connection string.
        """
        return AnyHttpUrl.build(
            scheme=self.SCHEME,
            username=self.USERNAME,
            password=self.PASSWORD,
            host=self.HOST,
            port=self.PORT,
        )


class PostgresConnection(DBConnection[PostgresDsn]):
    """
    Represents a connection to a Postgres database.

    Attributes:
        USERNAME (str): The username for the Postgres database.
        PASSWORD (str): The password for the Postgres database.
        HOST (str): The host of the Postgres database.
        PORT (int): [Optional] The port of the Postgres database. Default is 5432.
        DATABASE (str): The name of the Postgres database.

        CONN_STR (str): The connection string for the Postgres database, assembled from the
                                       other properties.
    """

    USERNAME: str
    PASSWORD: str
    HOST: str
    PORT: int = 5432
    DATABASE: str

    def _assemble_conn_str_value(self: Self) -> PostgresDsn:
        """
        Assembles a connection string representation for a Postgres database connection.

        Arguments:
            self (Self): A reference to the instance of PostgresConnection that this method will be called on.

        Returns:
            PostgresDsn: A Pydantic Postgres DSN model that represents the connection string.
        """
        return PostgresDsn.build(
            scheme="postgresql",
            username=self.USERNAME,
            password=self.PASSWORD,
            host=self.HOST,
            port=self.PORT,
            path=self.DATABASE or "",
        )


class RedisConnection(DBConnection[RedisDsn]):
    """
    Represents a Redis database connection string.

    Attributes:
        PASSWORD (str): The password for the Redis database.
        HOST (str): The host of the Redis database.
        PORT (int): [Optional] The port for the Redis database. Default is 6379.
        DATABASE (str): [Optional] The database index for the Redis database. Default is "0".
        USE_SSL (bool): [Optional] If true, the Redis service is assumed to use SSL. Default is True.

        CONN_STR (str): The connection string for the Redis service, assembled from the other properties.
    """

    PASSWORD: str | None = None
    HOST: str
    PORT: int = 6379
    DATABASE: str = "0"

    USE_SSL: bool = True

    def _assemble_conn_str_value(self: Self) -> RedisDsn:
        """
        Assembles a connection string representation for a Redis database connection.

        Arguments:
            self (Self): A reference to the instance of RedisConnection that this method will be called on.

        Returns:
            RedisDsn: A Pydantic Redis DSN model that represents the connection string.
        """
        scheme = "rediss" if self.USE_SSL else "redis"
        return RedisDsn.build(
            scheme=scheme,
            password=self.PASSWORD,
            host=self.HOST.split("://")[-1],
            port=self.PORT,
            path=self.DATABASE or "",
        )



"""This module contains all possible error codes used across microservices.

The intention is to avoid error codes duplication, standardize categories, provide messages and descriptions
through enums and dataclasses.

These error codes are not intended to be reused several times, and it's better
to create a separate error for every exception handling/raise. With that it's much easier to debug the application.

Classes:
    ErrorCodesKeeper: Dataclass for all error codes.

Attributes:
    ERROR_CODES: Constant entrypoint to all existing error codes.

"""

__all__ = [
    "ERROR_CODES",
    "ErrorCode",
    "ErrorCodesKeeper",
    "Prefix",
    "error_code_keeper",
]

from typing import Final

from kyotsu.logging.error_codes.dataclasses import ErrorCode, error_code_keeper
from kyotsu.logging.error_codes.definitions import ErrorCodesKeeper, Prefix

ERROR_CODES: Final[ErrorCodesKeeper] = ErrorCodesKeeper()
"""An entrypoint to the ✨error codes world✨. Keeps all error codes."""



"""
This module defines `ErrorCode` class and `error_code_keeper` decorator
to handle application errors efficiently.


Classes:
    ErrorCode: A dataclass representing a structured error code that includes several descriptions useful for
               different audiences such as end user, developers, and support staff.

Methods:
    error_code_keeper: A decorator that creates a class that contains all error codes.


Example:
    Example of creating your error codes and keepers.

    ```python
    from kyotsu.logging.error_codes import Prefix

    @error_code_keeper
    class AD_CErrorCodes:
        ZERO_DIVISION_ERROR: Final[ErrorCode] = ErrorCode(
            _code=1000,
            _prefix=Prefix.AD_C,
            _message="Divider can't be zero, please try another number",
            _dev_description="Handled ZeroDivisionError exception caused by user input.",
            _hp_description="User tried to divide by zero. The zero shouldn't be entered as a divider.",
        )

    @error_code_keeper
    class SCHErrorCodes:
        EXTERNAL_SERVICE_ERROR: Final[ErrorCode] = ErrorCode(
            1001,
            Prefix.SCH,
            _dev_description="EXTERNAL_SERVICE returned an error. Check service logs for more details.",
            _hp_description="Error occurred during services communication. Raise a request for the dev team.",
        )

    @error_code_keeper
    class MixedErrorCodes:
        AD_C: Final[AD_CErrorCodes] = AD_CErrorCodes()
        SCH: Final[SCHErrorCodes] = SCHErrorCodes()

    ERROR_CODES: Final[MixedErrorCodes] = MixedErrorCodes()
    ```
"""

__all__ = [
    "ErrorCode",
    "error_code_keeper",
]

import dataclasses
from collections.abc import Hashable
from enum import StrEnum
from typing import Self, TypeVar


@dataclasses.dataclass(
    slots=True,
    frozen=True,
    eq=True,
)
class ErrorCode:
    """This class is used to define an error code and use it later in the application.

    Attributes:
        _prefix: A prefix for categorization error.
        _code: A numeric code for the error, should be >= 1.
        message: An error message that can be displayed to the user.
                  If the error is not designed for the end user keep the default value.
        code: A property that should be used to obtain full error code. Combination of `_prefix` and `_code` attributes.


    Example:
        Fully initialized
        ```python3
            >>> error = ErrorCode("A", 1, "Some Error Message")
            >>> error
            ErrorCode(_prefix='A', _code=1, message='Some Error Message')
            >>> error.code
            'A0001'
            >>> str(error)
            'A0001: Some Error Message'

        ```

        Fallback to default values
        ```python3
        >>> error = ErrorCode("B", 11)
        >>> error
        ErrorCode(_prefix='B', _code=11, message='An unknown error has occurred')
        >>> error.code
        'B0011'
        >>> str(error)
        'B0011: An unknown error has occurred'

        ```

        The class is frozen and can't be modified after initialization.
        ```python3
        >>> error = ErrorCode("C", 2222)
        >>> error._code = 2
        Traceback (most recent call last):
        dataclasses.FrozenInstanceError: ...

        ```

        The `ErrorCode` is [Hashable][collections.abc.Hashable],
        therefore you can use any instance as a dictionary key,
        if you need it for some reason.
        ```python3
        >>> error = ErrorCode("D", 333)
        >>> d = {error: error.code}
        >>> d[error]
        'D0333'

        ```
    """

    _prefix: str | StrEnum
    _code: int
    message: str = "An unknown error has occurred"

    @property
    def code(self: Self) -> str:
        """Concatenate `self._prefix` and `self._code`."""
        return f"{self._prefix}{self._code:0>4}"

    def __str__(self: Self) -> str:
        return f"{self.code}: {self.message}"


_T = TypeVar("_T")

t = Hashable


def error_code_keeper(__cls: type[_T]) -> type[_T]:
    """A decorator that creates a dataclass with `init=False` and `frozen=True`.

    Should be used to create error keepers and store [kyotsu.logging.error_codes.ErrorCode][]s.

    Attributes:
        __cls (type[_T]): Decorated class that should be augmented with dataclass methods

    Returns:
        A dataclass with `init=False` and `frozen=True`.

    Example:
        Using `@error_code_keeper` decorator to create new keeper.

        ```python3
        >>> from kyotsu.logging.error_codes.dataclasses import error_code_keeper, ErrorCode
        >>> from typing import Final

        >>> @error_code_keeper
        ... class BErrorKeeper:
        ...     UNEXPECTED_FOR_B: Final[ErrorCode] = ErrorCode("B", 1, "Unknown error in B context")

        >>> @error_code_keeper
        ... class AErrorKeeper:
        ...     UNEXPECTED_FOR_A: Final[ErrorCode] = ErrorCode("A", 1, "Unknown error in A context")

        >>> @error_code_keeper
        ... class ErrorKeeper:
        ...     GENERIC_UNEXPECTED: Final[ErrorCode] = ErrorCode("UN", 1)
        ...     B: Final[BErrorKeeper] = BErrorKeeper()
        ...     A: Final[AErrorKeeper] = AErrorKeeper()

        >>> ERROR_CODES = ErrorKeeper()

        >>> ERROR_CODES.GENERIC_UNEXPECTED
        ErrorCode(_prefix='UN', _code=1, message='An unknown error has occurred')
        >>> ERROR_CODES.B.UNEXPECTED_FOR_B
        ErrorCode(_prefix='B', _code=1, message='Unknown error in B context')
        >>> ERROR_CODES.A.UNEXPECTED_FOR_A
        ErrorCode(_prefix='A', _code=1, message='Unknown error in A context')

        ```
    """
    return dataclasses.dataclass(init=False, frozen=True)(__cls)



from collections.abc import Mapping
from typing import Any

from pydantic import BaseModel, ConfigDict, Field
from pydantic.alias_generators import to_camel


class ErrorModel(BaseModel):
    code: str = Field(
        ...,
        title="Error code",
        description="Error code according to error codes DB.",
        examples=["UN0001", "UN0002"],
    )
    details: str | None = Field(
        None,
        title="Error details",
        description="Error details (message) available to end user.",
        examples=["Unknown error has occurred", "Request Validation Error"],
    )
    ray_id: str | None = Field(
        None,
        title="Error Trace ID",
        description="Trace ID of the span where error occurred.",
        examples=["dd647833-339c-4080-b7a1-bebfbd2365d0", "a4f5fab3-abc0-48de-a6ff-c142fb27643b"],
        json_schema_extra={"format": "uuid"},
    )
    extra: str | Mapping[str, Any] | list[Any] | None = Field(
        None,
        title="Dev Error Details",
        description="Details that can be useful for developers.",
        examples=[
            {"exception": "ZeroDivisionError"},
            {
                "validationDetails": [
                    {
                        "type": "datetime_from_date_parsing",
                        "loc": ["body", "expires_at"],
                        "msg": (
                            "Input should be a valid datetime or date,"
                            "unexpected extra characters at the end of the input"
                        ),
                        "input": "2024-03-05TT23:35:00.323Z",
                        "ctx": {"error": "unexpected extra characters at the end of the input"},
                        "url": "https://errors.pydantic.dev/2.6/v/datetime_from_date_parsing",
                    },
                    {
                        "type": "missing",
                        "loc": ["body", "start_at"],
                        "msg": "Field required",
                        "input": {"key1": "value1", "key2": "value2"},
                        "url": "https://errors.pydantic.dev/2.6/v/missing",
                    },
                ],
            },
        ],
    )

    model_config = ConfigDict(alias_generator=to_camel, populate_by_name=True)



__all__ = [
    "AsyncSessionMaker",
    "Base",
    "SessionMaker",
    "TableNameMixin",
    "TimeLoggedMixinDDL",
    "TimeLoggedMixinORM",
    "UUIDMixin",
]

from kyotsu.db.sqlalchemy.base import Base
from kyotsu.db.sqlalchemy.mixins import TableNameMixin, TimeLoggedMixinDDL, TimeLoggedMixinORM, UUIDMixin
from kyotsu.db.sqlalchemy.session_maker import AsyncSessionMaker, SessionMaker



"""
This module contains the unit tests for classes and functions within
the `kyotsu.logging.opensearch.processors` module.
"""

import json

import pytest

from kyotsu.logging.opensearch.processors import SafeJSONSerializer


class NonSerializableObject:
    """
    A simple mockup class that represents an object that cannot be serialized with the standard JSON serializer.
    """

    def __repr__(self) -> str:
        return "NonSerializableObject()"


# Instance of SafeJSONSerializer
serializer = SafeJSONSerializer()

# Test data for parameterization
test_data = [
    pytest.param({"key": NonSerializableObject(), "other_key": 123}, ["key"], id="object_at_root"),
    pytest.param(
        {"key": {"nested_key": NonSerializableObject(), "nested_key2": 456}},
        ["key", "nested_key"],
        id="object_in_nested_dict",
    ),
    pytest.param({"key": [1, "string", NonSerializableObject(), 4.5]}, ["key", 2], id="object_in_list"),
]


@pytest.mark.parametrize(("data", "path"), test_data)
def test_safe_json_serializer(data, path):
    """
    Test that the SafeJSONSerializer correctly replaces non-serializable objects with a string representation.

    Args:
        data (dict): The data to be serialized. It contains a non-serializable object in a location specified by path.
        path (list): The path to the non-serializable object in the data.

    Raises:
        AssertionError: If the SafeJSONSerializer doesn't replace the non-serializable object correctly.
    """

    serialized = serializer.dumps(data)
    loaded = json.loads(serialized)

    # Extract element at `path`.
    element = loaded
    for p in path:
        element = element[p]

    assert element == f"<<nonserializable: {NonSerializableObject()!r}>>"



__all__ = [
    "LoggingGroup",
]

from enum import StrEnum, auto
from typing import Any


class LoggingGroup(StrEnum):
    """
    Predefined logging groups.

    This Enum is designed to generate value automatically, by replacing `_` in name wih `-` and making it lowercase.
    Always prioritize `auto()` over manual value assigment.

    Attributes:

        DEFAULT: Default logging group. Reference to APPLICATION.
        APPLICATION: Application logging group. Generic log group. Everything uncategorized should go here.
        PRODUCT: Product logging group. Used to mark product logs that are used by product analytics.
        NETWORK: Network logging group. Used to mark logs related to external services.
        DATABASE: Database logging group. Used to mark logs related to database operations.
        PERFORMANCE: Performance logging group. Used to mark logs related to performance metrics.
    """

    @staticmethod
    def _generate_next_value_(name: str, *args: Any) -> str:
        """Build group name by replacing `_` in name with `-` and making it lowercase."""
        return name.replace("_", "-").lower()

    DEFAULT = "application"
    APPLICATION = auto()
    PRODUCT = auto()
    NETWORK = auto()
    DATABASE = auto()
    PERFORMANCE = auto()



__all__ = [
    "Base",
]

from typing import Self

from sqlalchemy.ext.asyncio import AsyncAttrs
from sqlalchemy.orm import DeclarativeBase

from kyotsu.db.sqlalchemy.mixins import TableNameMixin


class Base(AsyncAttrs, TableNameMixin, DeclarativeBase):
    def __repr__(self: Self) -> str:
        column_repr = (f"{c.key}={getattr(self, c.key)!r}" for c in self.__table__.columns)
        return f"{self.__class__.__name__}({", ".join(column_repr)})"



"""
This module contains base models for defining application settings involving HTTP and gRPC services.

These models should be used in a nested manner with [`BaseSettings`][kyotsu.config.base.BaseSettings] class.

Example:
    ```python3
    >>> from kyotsu.config import BaseSettings, HttpService, GrpcService, HttpGrpcService

    >>> class Settings(BaseSettings):
    ...     EXTERNAL_HTTP_SERVICE_NAME: HttpService
    ...     EXTERNAL_GRPC_SERVICE_NAME: GrpcService
    ...     EXTERNAL_HTTP_GRPC_SERVICE_NAME: HttpGrpcService

    ```

In the above example, `EXTERNAL_HTTP_SERVICE_NAME`, `EXTERNAL_GRPC_SERVICE_NAME`, and `EXTERNAL_HTTP_GRPC_SERVICE_NAME`
will contain the configuration for each respective service type.

Classes:
    HttpService: Defines and validates the HTTP_URL attribute representing the URL for HTTP communication
                 with a service.
    GrpcService: Defines and validates the GRPC_HOST and GRPC_PORT attributes required for GRPC communication
                 with a service.
    HttpGrpcService: A combination of HttpService and GrpcService, for services that offer both
                     HTTP and GRPC interfaces.
"""

__all__ = [
    "GrpcService",
    "HttpGrpcService",
    "HttpService",
]

from typing import Annotated

from pydantic import BaseModel, HttpUrl


class HttpService(BaseModel):
    """
    This class represents the standard attributes for a service that interacts via HTTP.

    Attributes:
        HTTP_URL: The URL for the HTTP service.
    """

    HTTP_URL: Annotated[str, HttpUrl]


class GrpcService(BaseModel):
    """
    This class represents the standard attributes for a service that interacts via gRPC.

    Attributes:
        GRPC_HOST: The host for the gRPC service.
        GRPC_PORT: The port for the gRPC service.
    """

    GRPC_HOST: str
    GRPC_PORT: str


class HttpGrpcService(HttpService, GrpcService):
    """
    A combined service class that possesses both HTTP and gRPC attributes.

    Attributes:
        HTTP_URL: The URL for the HTTP service.
        GRPC_HOST: The host for the gRPC service.
        GRPC_PORT: The port for the gRPC service.
    """



__all__ = [
    "error_telemetry_adder",
]

from structlog.processors import _figure_out_exc_info
from structlog.typing import EventDict, WrappedLogger


def error_telemetry_adder(_: WrappedLogger, __: str, event_dict: EventDict) -> EventDict:
    exc_info = _figure_out_exc_info(event_dict.get("exc_info"))

    if not exc_info or all(e is None for e in exc_info):
        return event_dict

    try:
        telemetry = exc_info[1].telemetry  # type: ignore[attr-defined] # We are handling AttributeError
    except AttributeError:
        return event_dict

    try:
        event_dict["error_details"] |= telemetry
    except KeyError:
        event_dict["error_details"] = telemetry

    return event_dict



"""
This module provides the models and mix-ins necessary for configuring and maintaining the connections
to various databases across the application.

The configuration for different types of database connections are defined and standardized
in this package, ensuring the consistency and maintainability of the codebase.

Modules:
    models: Defines the models for various types of database connections used across the application.
    _types: Contains inner misc types for mixins and models.

When needing to configure a new database connection or reuse an existing common configuration across multiple services,
these modules should be the first place to look.
"""

__all__ = [
    "AmqpConnection",
    "DBConnection",
    "OpenSearchNoCertConnection",
    "PostgresConnection",
    "RedisConnection",
]

from .models import AmqpConnection, DBConnection, OpenSearchNoCertConnection, PostgresConnection, RedisConnection



__all__ = [
    "JsonAlias",
]

from collections.abc import Mapping, Sequence
from typing import TypeAlias

JsonBasics: TypeAlias = str | int | float | bool | None  # noqa: UP040 # MyPy 1.8.0 does not support type keyword yet
JsonAlias: TypeAlias = JsonBasics | Sequence["JsonAlias"] | Mapping[str, "JsonAlias"]  # noqa: UP040 # MyPy 1.8.0 does not support type keyword yet



__all__ = [
    "AsyncTelemetryClient",
    "TelemetryClient",
    "generate_ray_id",
    "log_span_started",
    "nest_span_id",
    "parent_span_id",
    "span_id",
    "start_telemetry",
    "trace_id",
]

from .clients import AsyncTelemetryClient, TelemetryClient
from .contextvars import parent_span_id, span_id, trace_id
from .utils import (
    generate_ray_id,
    log_span_started,
    nest_span_id,
    start_telemetry,
)



from .metadata import (
    AppMetaData,
    MetaDataContact,
    MetaDataLicense,
    MetaDataServer,
    MetaDataTag,
    MetaDataTagExternalDocs,
    OpenAPIMetaData,
)
from .routes import ContentExample, RequestExample, ResponseExample, RouteData



__all__ = [
    "log_exception",
    "transform_exceptions",
]

from asyncio import iscoroutinefunction
from collections.abc import Awaitable, Callable, Iterable
from types import TracebackType
from typing import Self

from kyotsu.logging.error_codes import ERROR_CODES
from kyotsu.logging.exceptions import BaseKyotsuError, ClientKyotsuError
from kyotsu.utils.contextmanagers import SyncAsyncContextDecorator


class transform_exceptions(SyncAsyncContextDecorator):  # noqa: N801 # snake_case to fit decorator and contex manager convention
    def __init__(
        self,
        *,
        skip_transform: Iterable[type[BaseException]] | None = None,
        transform_only: Iterable[type[BaseException]] | None = None,
    ) -> None:
        self.skip_transform: Iterable[type[BaseException]] = (
            skip_transform if skip_transform is not None else {BaseKyotsuError, ClientKyotsuError}
        )
        self.transform_only: Iterable[type[BaseException]] = transform_only or set()

    def __enter__(self: Self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool:
        if exc_type:
            if exc_type in self.transform_only or (not self.transform_only and exc_type not in self.skip_transform):
                msg = f"Unhandled {exc_type.__name__} occurred"
                raise BaseKyotsuError(
                    error_code=ERROR_CODES.UN.UNHANDLED_EXCEPTION,
                ) from exc_val
            return False
        return True


class log_exception(SyncAsyncContextDecorator):  # noqa: N801 # snake_case to fit decorator and contex manager convention
    def __init__(
        self,
        log: Callable[..., None] | Callable[..., Awaitable[None]],
        *,
        skip_log: Iterable[type[BaseException]] | None = None,
        log_only: Iterable[type[BaseException]] | None = None,
    ) -> None:
        self.log = log
        self.skip_log: Iterable[type[BaseException]] = skip_log or set()
        self.log_only: Iterable[type[BaseException]] = log_only or set()

    def __enter__(self: Self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool:
        if iscoroutinefunction(self.log):
            msg = f"Async log method '{self.log}' can't be used in sync context manager."
            raise BaseKyotsuError(error_code=ERROR_CODES.KTS.ASYNC_LOGGER_IN_SYNC_CONTEXT)
        if exc_type:
            if exc_type in self.log_only or (not self.log_only and exc_type not in self.skip_log):
                self.log(str(exc_val), exc_info=(exc_type, exc_val, exc_tb), extra=getattr(exc_val, "extra", None))
            return False
        return True

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool:
        if exc_type:
            if exc_type in self.log_only or (not self.log_only and exc_type not in self.skip_log):
                if iscoroutinefunction(self.log):
                    await self.log(
                        str(exc_val),
                        exc_info=(exc_type, exc_val, exc_tb),
                        extra=getattr(exc_val, "extra", None),
                    )
                else:
                    self.log(str(exc_val), exc_info=(exc_type, exc_val, exc_tb), extra=getattr(exc_val, "extra", None))
            return False
        return True



"""Danger:
Only modify these classes if you are completely certain of the implications, as numerous core settings
in external services are dependent on them.
"""

from .base import BaseSettings
from .services import GrpcService, HttpGrpcService, HttpService



__all__ = [
    "BaseKyotsuError",
    "ClientKyotsuError",
]

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Self, TypedDict

from kyotsu.logging.groups import LoggingGroup
from kyotsu.logging.telemetry.contextvars import parent_span_id, span_id, trace_id
from kyotsu.utils import JsonAlias

if TYPE_CHECKING:
    from kyotsu.logging.error_codes.dataclasses import ErrorCode


class ErrorModelMapping(TypedDict, total=False):
    code: str
    details: str
    ray_id: str
    extra: Mapping[str, Any]


class BaseKyotsuError(Exception):
    error_code: "ErrorCode"
    status_code: int
    group: LoggingGroup | None
    telemetry: dict[str, str | None]
    http_headers: dict[str, str]
    extra: dict[str, JsonAlias]

    def __init__(
        self: Self,
        error_code: "ErrorCode",
        status_code: int = 500,
        group: LoggingGroup | None = None,
        *,
        _trace_id: str | None = None,
        _span_id: str | None = None,
        _parent_span_id: str | None = None,
        _http_headers: dict[str, str] | None = None,
        **extra: JsonAlias,
    ):
        self.error_code = error_code
        self.group = group
        self.status_code = status_code
        self.telemetry = {
            "trace_id": _trace_id or trace_id.get(),
            "span_id": _span_id or span_id.get(),
            "parent_span_id": _parent_span_id or parent_span_id.get(),
        }
        self.http_headers = {} if _http_headers is None else _http_headers
        self.extra = extra
        super().__init__(str(self.error_code))

    def __repr__(self: Self) -> str:
        base_str = f"{self.__class__.__name__}({self.error_code!r}"
        if self.status_code != 500:
            base_str += f", status_code={self.status_code}"
        if self.group:
            base_str += f", group={self.group}"
        if self.telemetry:
            k_v = ", ".join(f"_{k}={v}" for k, v in self.telemetry.items() if v)
            base_str += f", {k_v}"
        if self.http_headers:
            base_str += f", _http_headers={self.http_headers}"
        if self.extra:
            k_v = ", ".join(f"{k}={v!r}" for k, v in self.extra.items())
            base_str += f", {k_v}"
        return f"{base_str})"

    @property
    def response_data(self: Self) -> tuple[int, ErrorModelMapping, dict[str, str]]:
        body = ErrorModelMapping(
            code=self.error_code.code,
            details=self.error_code.message,
        )
        if self.telemetry["trace_id"]:
            body["ray_id"] = self.telemetry["trace_id"]
        return self.status_code, body, self.http_headers


class ClientKyotsuError(BaseKyotsuError):
    @property
    def response_data(self: Self) -> tuple[int, ErrorModelMapping, dict[str, str]]:
        status_code, body, http_headers = super().response_data
        if self.extra:
            body["extra"] = self.extra
        return status_code, body, http_headers



__all__ = [
    "SpanKind",
    "SpanStatus",
]

from enum import StrEnum, unique


@unique
class SpanKind(StrEnum):
    """
    https://opentelemetry.io/docs/concepts/signals/traces/#span-kind
    """

    CLIENT = "CLIENT"
    SERVER = "SERVER"
    INTERNAL = "INTERNAL"
    PRODUCER = "PRODUCER"
    CONSUMER = "CONSUMER"


@unique
class SpanStatus(StrEnum):
    """
    https://opentelemetry.io/docs/concepts/signals/traces/#span-status
    """

    UNSET = "Unset"
    OK = "Ok"
    ERROR = "Error"



from kyotsu.logging.config import configure_logger
from kyotsu.logging.error_codes import ERROR_CODES, ErrorCodesKeeper
from kyotsu.logging.exceptions import BaseKyotsuError, ClientKyotsuError
from kyotsu.logging.telemetry import (
    AsyncTelemetryClient,
    TelemetryClient,
    generate_ray_id,
    log_span_started,
    nest_span_id,
    parent_span_id,
    span_id,
    start_telemetry,
    trace_id,
)



"""
The `kyotsu.config.base` module provides a base settings configuration class for the application.
This class ensures a uniform way to manage configurations including environment variables across the
entire application, contributing to the maintenance and consistency of the codebase.

Classes:
    BaseSettings: An abstract class that encapsulates the base configuration settings for models using
                  the Pydantic settings configuration approach.
"""

__all__ = [
    "BaseSettings",
]

from pydantic_settings import BaseSettings as PydanticSettings
from pydantic_settings import SettingsConfigDict


class BaseSettings(PydanticSettings):
    """
    A base settings configuration class that simplifies environment-based settings for Pydantic models.

    It includes environmental settings specified as '__'-separated nested properties and the encoding
    for .env files is 'utf-8'.

    Attributes:
        IS_DEV: A boolean flag indicating whether the application is running in development mode.
        model_config (SettingsConfigDict): Configuration dictionary for the settings model.
                                           It sets the delimiter and encoding standard for environment variables.
    """

    IS_DEV: bool = False

    model_config = SettingsConfigDict(
        extra="allow",
        env_nested_delimiter="__",
        env_file_encoding="utf-8",
    )



import abc
from asyncio import iscoroutinefunction
from collections.abc import Awaitable, Callable, Sequence
from typing import TYPE_CHECKING, Self

import structlog.stdlib
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import RequestValidationError, ResponseValidationError
from fastapi.responses import JSONResponse
from starlette.middleware.base import BaseHTTPMiddleware

from kyotsu.fastapi.models import ErrorModel
from kyotsu.logging.error_codes import ERROR_CODES
from kyotsu.logging.exceptions import BaseKyotsuError, ClientKyotsuError
from kyotsu.logging.telemetry.constants import SpanKind
from kyotsu.logging.telemetry.utils import generate_ray_id, log_span_started, start_telemetry
from kyotsu.logging.utils import log_exception, transform_exceptions

if TYPE_CHECKING:
    from fastapi import FastAPI, Request, Response


class BaseLoggingMiddleware(abc.ABC):
    def __init__(self: Self, log_method: Callable[..., None] | Callable[..., Awaitable[None]]):
        self.log_method = log_method

    @abc.abstractmethod
    async def __call__(
        self: Self,
        request: "Request",
        call_next: Callable[["Request"], Awaitable["Response"]],
    ) -> "Response": ...


async def handle_errors(request: "Request", call_next: Callable[["Request"], Awaitable["Response"]]) -> "Response":
    try:
        response = await call_next(request)
    except BaseKyotsuError as exc:
        status_code, body, headers = exc.response_data
        return JSONResponse(
            status_code=status_code,
            content=jsonable_encoder(
                ErrorModel(**body).model_dump(exclude_unset=True),
            ),
            headers=headers,
        )
    return response


class StartTelemetryMiddlewareBase(BaseLoggingMiddleware):
    async def __call__(
        self: Self,
        request: "Request",
        call_next: Callable[["Request"], Awaitable["Response"]],
    ) -> "Response":
        h_trace_id = request.headers.get("X-Trace-ID", generate_ray_id())
        h_span_id = request.headers.get("X-Span-ID", generate_ray_id())
        async with (
            start_telemetry(_trace_id=h_trace_id, _span_id=h_span_id),
        ):
            if request.url.path != "/ping":
                async with log_span_started(self.log_method, span_kind=SpanKind.SERVER):
                    result = await call_next(request)
            else:
                result = await call_next(request)
        return result  # noqa: RET504 # Can't move return into ctx manager: https://github.com/python/mypy/issues/7577#issuecomment-535991883


async def exceptions_transformer_middleware(
    request: "Request",
    call_next: Callable[["Request"], Awaitable["Response"]],
) -> "Response":
    try:
        async with transform_exceptions(
            skip_transform={
                BaseKyotsuError,  # Already transformed
                ClientKyotsuError,  # Already transformed
                ResponseValidationError,  # Will be transformed manually lower
                RequestValidationError,  # Will be transformed manually lower
            },
        ):
            response = await call_next(request)
    except RequestValidationError as exc:
        msg = ("Unprocessable entity in request",)
        raise ClientKyotsuError(
            ERROR_CODES.UN.REQUEST_VALIDATION_ERROR,
            status_code=422,
            validationDetails=exc.errors(),
        ) from exc
    except ResponseValidationError as exc:
        msg = ("Unprocessable entity in response",)
        raise BaseKyotsuError(
            ERROR_CODES.UN.RESPONSE_VALIDATION_ERROR,
            status_code=500,
            validation_details=exc.errors(),  # We want to log details, but avoid adding them to response
        ) from exc
    return response


class ExceptionLoggerMiddleware(BaseLoggingMiddleware):
    async def __call__(
        self: Self,
        request: "Request",
        call_next: Callable[["Request"], Awaitable["Response"]],
    ) -> "Response":
        async with log_exception(self.log_method):
            result = await call_next(request)
        return result  # noqa: RET504 # Can't move return into ctx manager: https://github.com/python/mypy/issues/7577#issuecomment-535991883


class PathLogMiddleware(BaseLoggingMiddleware):
    async def __call__(
        self: Self,
        request: "Request",
        call_next: Callable[["Request"], Awaitable["Response"]],
    ) -> "Response":
        result = await call_next(request)
        status_code = result.status_code
        method = request.method
        path = request.url.path
        if path != "/ping":
            if iscoroutinefunction(self.log_method):
                await self.log_method("Processed request", method=method, path=path, status_code=status_code)
            else:
                self.log_method("Processed request", method=method, path=path, status_code=status_code)
        return result


def get_middleware_stack(
    log_info_method: Callable[..., None] | Callable[..., Awaitable[None]],
    log_error_method: Callable[..., None] | Callable[..., Awaitable[None]],
) -> list[Callable[["Request", Callable[["Request"], Awaitable["Response"]]], Awaitable["Response"]]]:
    return [
        handle_errors,
        StartTelemetryMiddlewareBase(log_info_method),
        ExceptionLoggerMiddleware(log_error_method),
        exceptions_transformer_middleware,
        PathLogMiddleware(log_info_method)
    ]


def apply_middleware_stack(
    app: "FastAPI",
    middlewares_stack: Sequence[
        Callable[["Request", Callable[["Request"], Awaitable["Response"]]], Awaitable["Response"]]
    ],
) -> None:
    for middleware in reversed(middlewares_stack):
        app.add_middleware(BaseHTTPMiddleware, dispatch=middleware)



__all__ = [
    "Prefix",
]

from enum import StrEnum, auto, unique
from typing import Any


@unique
class Prefix(StrEnum):
    """Enum that keeps prefixes for [kyotsu.logging.error_codes.ErrorCode][]s.

    This Enum is designed to generate value automatically, by replacing `_` in name wih `-`.
    Always prioritize `auto()` over manual value assigment.

    Attributes:
        UN: Uncategorized Errors
        KTS: Kyotsu Runtime Errors
        SCH: Scheduler Errors
    """

    @staticmethod
    def _generate_next_value_(name: str, *args: Any) -> str:
        """Build prefix by replacing `_` in name with `-`."""
        return name.replace("_", "-")

    UN = auto()
    KTS = auto()
    SCH = auto()



__all__ = [
    "OpenSearchLogProcessor",
]

import datetime
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Self

from opensearchpy import JSONSerializer, OpenSearch
from structlog.typing import EventDict, WrappedLogger

if TYPE_CHECKING:
    from kyotsu.config.databases import OpenSearchNoCertConnection

from ..groups import LoggingGroup


class SafeJSONSerializer(JSONSerializer):
    def default(self: Self, data: Any) -> Any:
        try:
            return super().default(data)
        except TypeError:
            return f"<<nonserializable: {data!r}>>"


class OpenSearchLogProcessor:
    def __init__(
        self: Self,
        os_settings: "OpenSearchNoCertConnection",
        log_group_index_mapping: Mapping[str, str],
        *,
        default_index: str,
    ):
        self.conn = OpenSearch(**os_settings.CONN_DICT, serializer=SafeJSONSerializer())
        self.log_group_index_mapping = log_group_index_mapping
        self.default_index = default_index

    def __call__(self: Self, _: WrappedLogger, __: str, event_dict: EventDict) -> EventDict:
        log_group = event_dict.get("group", LoggingGroup.APPLICATION)
        index_name = self.log_group_index_mapping.get(log_group, self.default_index)
        index = f"{index_name}-{datetime.datetime.now(datetime.UTC)}"
        self.conn.index(index=index, body=event_dict)
        return event_dict



"""This module contains keeper class for uncategorized errors."""

__all__ = [
    "KyotsuKeeper",
]

from typing import Final

from kyotsu.logging.error_codes.dataclasses import ErrorCode, error_code_keeper

from ._constants import Prefix


@error_code_keeper
class KyotsuKeeper:
    """Collection of kyotsu errors.

    Danger:
        Do not add anything to this category unless you are absolutely sure in your actions.
        # ToDo: Add link to recipes
        See: [Using Error Prefix]()

    Attributes:
        ASYNC_LOGGER_IN_SYNC_CONTEXT: Async log method was passed as callable to sync context manager.
    """

    ASYNC_LOGGER_IN_SYNC_CONTEXT: Final[ErrorCode] = ErrorCode(
        Prefix.KTS,
        1
    )



__all__ = [
    "generate_ray_id",
    "log_span_started",
    "nest_span_id",
    "start_telemetry",
]

import uuid
from asyncio import iscoroutinefunction
from collections.abc import Awaitable, Callable
from types import TracebackType
from typing import Self

import structlog.contextvars

from kyotsu.logging.error_codes import ERROR_CODES
from kyotsu.logging.exceptions import BaseKyotsuError
from kyotsu.logging.telemetry.constants import SpanKind, SpanStatus
from kyotsu.logging.telemetry.contextvars import parent_span_id, span_id, trace_id
from kyotsu.utils.contextmanagers import SyncAsyncContextDecorator


def generate_ray_id() -> str:
    return str(uuid.uuid4())


class start_telemetry(SyncAsyncContextDecorator):  # noqa: N801 # snake_case to fit decorator and contex manager convention
    def __init__(
        self,
        _trace_id: str | None = None,
        _span_id: str | None = None,
        _parent_span_id: str | None = None,
    ) -> None:
        self._trace_id = _trace_id
        self._span_id = _span_id
        self._parent_span_id = _parent_span_id

    def __enter__(self: Self) -> Self:
        self.trace_id_token = trace_id.set(self._trace_id or generate_ray_id())
        self.span_id_token = span_id.set(self._span_id or generate_ray_id())
        self.parent_span_id_token = None
        if self._parent_span_id:
            self.parent_span_id_token = parent_span_id.set(self._parent_span_id)

        self._structlog_tokens = structlog.contextvars.bind_contextvars(
            trace_id=trace_id.get(),
            span_id=span_id.get(),
            parent_span_id=parent_span_id.get(),
        )

        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool:
        trace_id.reset(self.trace_id_token)
        span_id.reset(self.span_id_token)

        if self.parent_span_id_token:
            parent_span_id.reset(self.parent_span_id_token)

        structlog.contextvars.reset_contextvars(**self._structlog_tokens)

        return not exc_type


class nest_span_id(SyncAsyncContextDecorator):  # noqa: N801 # snake_case to fit decorator and contex manager convention
    def __enter__(self: Self) -> Self:
        self.parent_span_id_token = parent_span_id.set(span_id.get())
        self.span_id_token = span_id.set(generate_ray_id())
        self._structlog_tokens = structlog.contextvars.bind_contextvars(
            span_id=span_id.get(),
            parent_span_id=parent_span_id.get(),
        )

        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool:
        parent_span_id.reset(self.parent_span_id_token)
        span_id.reset(self.span_id_token)

        structlog.contextvars.reset_contextvars(**self._structlog_tokens)

        return not exc_type


class log_span_started(SyncAsyncContextDecorator):  # noqa: N801 # snake_case to fit decorator and contex manager convention
    def __init__(
        self,
        log: Callable[..., None] | Callable[..., Awaitable[None]],
        span_kind: SpanKind,
    ):
        self.log = log
        self.span_kind = span_kind

    def __enter__(self: Self) -> Self:
        if iscoroutinefunction(self.log):
            msg = f"Async log method '{self.log}' can't be used in sync context manager."
            raise BaseKyotsuError(ERROR_CODES.KTS.ASYNC_LOGGER_IN_SYNC_CONTEXT)
        self.log("Span Started", span_kind=self.span_kind, span_status=SpanStatus.UNSET)
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool:
        if iscoroutinefunction(self.log):
            msg = f"Async log method '{self.log}' can't be used in sync context manager."
            raise BaseKyotsuError(ERROR_CODES.KTS.ASYNC_LOGGER_IN_SYNC_CONTEXT)
        if exc_type:
            status, supress = SpanStatus.ERROR, False
        else:
            status, supress = SpanStatus.OK, True
        self.log("Span Ended", span_kind=self.span_kind, span_status=status)
        return supress

    async def __aenter__(self: Self) -> Self:
        if iscoroutinefunction(self.log):
            await self.log("Span Started", span_kind=self.span_kind, span_status=SpanStatus.UNSET)
        else:
            self.log("Span Started", span_kind=self.span_kind, span_status=SpanStatus.UNSET)
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool:
        if exc_type:
            status, supress = SpanStatus.ERROR, False
        else:
            status, supress = SpanStatus.OK, True
        if iscoroutinefunction(self.log):
            await self.log("Span Ended", span_kind=self.span_kind, span_status=status)
        else:
            self.log("Span Ended", span_kind=self.span_kind, span_status=status)
        return supress



__all__ = [
    "AppMetaData",
    "MetaDataContact",
    "MetaDataLicense",
    "MetaDataServer",
    "MetaDataTag",
    "MetaDataTagExternalDocs",
    "OpenAPIMetaData",
]

from dataclasses import asdict, dataclass, field
from typing import Self

from pydantic import BaseModel

from kyotsu.utils import JsonAlias

from .routes import ContentExample, ResponseExample


@dataclass
class MetaDataContact:
    name: str | None = None
    url: str | None = None
    email: str | None = None


@dataclass
class MetaDataLicense:
    name: str
    url: str | None = None


@dataclass
class MetaDataTagExternalDocs:
    url: str
    description: str | None = None


@dataclass(frozen=True)
class MetaDataTag:
    name: str
    description: str | None = None
    externalDocs: MetaDataTagExternalDocs | None = None  # noqa: N815 # following metadata dict case


@dataclass
class MetaDataServer:
    url: str
    description: str | None = None


@dataclass
class OpenAPIMetaData:
    title: str
    version: str
    description: str | None = None
    terms_of_service: str | None = None
    contact: MetaDataContact | None = None
    license_info: MetaDataLicense | None = None

    tags: list[MetaDataTag] | None = None
    servers: list[MetaDataServer] | None = None

    @property
    def unpack(self: Self) -> dict[str, JsonAlias]:
        return asdict(self)


def _get_default_responses() -> list[ResponseExample]:
    from kyotsu.fastapi.models import ErrorModel

    return [
        ResponseExample(
            status_code="4XX",
            description="Client Error",
            model=ErrorModel,
            examples=[
                ContentExample(
                    name="Validation Error",
                    value={
                        "code": "UN0002",
                        "details": "An unknown error occurred",
                        "rayId": "dd647833-339c-4080-b7a1-bebfbd2365d0",
                        "extra": {
                            "validationDetails": [
                                {
                                    "type": "datetime_from_date_parsing",
                                    "loc": ["body", "expires_at"],
                                    "msg": (
                                        "Input should be a valid datetime or date, "
                                        "unexpected extra characters at the end of the input"
                                    ),
                                    "input": "2024-03-05TT23:35:00.323Z",
                                    "ctx": {"error": "unexpected extra characters at the end of the input"},
                                    "url": "https://errors.pydantic.dev/2.6/v/datetime_from_date_parsing",
                                },
                                {
                                    "type": "missing",
                                    "loc": ["body", "start_at"],
                                    "msg": "Field required",
                                    "input": {"key1": "value1", "key2": "value2"},
                                    "url": "https://errors.pydantic.dev/2.6/v/missing",
                                },
                            ],
                        },
                    },
                ),
                ContentExample(
                    name="Not Found Error",
                    value={
                        "code": "AB1234",
                        "details": "Instance with id 123 not found",
                        "rayId": "255a55a3-7492-45a6-b40f-27e1e2e460b4",
                    },
                ),
            ],
        ),
        ResponseExample(
            status_code="5XX",
            description="Server Error",
            model=ErrorModel,
            examples=[
                ContentExample(
                    name="Internal Server Error",
                    value={
                        "code": "UN0001",
                        "details": "An unknown error occurred",
                        "rayId": "a4f5fab3-abc0-48de-a6ff-c142fb27643b",
                        "extra": {"exception": "ZeroDivisionError"},
                    },
                ),
            ],
        ),
    ]


@dataclass
class AppMetaData:
    openapi_url: str | None = "/openapi.json"
    docs_url: str | None = "/docs"
    redoc_url: str | None = "/redoc"
    _responses: list[ResponseExample] | None = field(default_factory=_get_default_responses)
    responses: dict[str, dict[str, JsonAlias | type[BaseModel]]] | None = field(init=False)

    @staticmethod
    def _get_response_example(example: ResponseExample) -> dict[str, JsonAlias | type[BaseModel]]:
        result: dict[str, JsonAlias | type[BaseModel]] = {"description": example.description}
        if example.model:
            result["model"] = example.model
        if example.examples:
            result["content"] = {
                "application/json": {
                    "examples": {content.name: {"value": content.value} for content in example.examples},
                },
            }
        return result

    def __post_init__(self: Self) -> None:
        if self._responses is None:
            self.responses = None
        else:
            self.responses = {
                response.status_code: self._get_response_example(response) for response in self._responses
            }

    @property
    def unpack(self: Self) -> dict[str, dict[int | str, JsonAlias | BaseModel]]:
        result = asdict(self)
        del result["_responses"]
        return result



from pydantic import BaseModel, RedisDsn


class CeleryConfig(BaseModel):
    CELERY_CONCURRENCY: int
    CELERY_PREFETCH_MULTIPLIER: int


class CeleryReadBeatConfig(BaseModel):
    """Represents configuration for celery-redbeat.

    By default, celery-redbeat will configure itself using celery settings,
    so use this class and override celery settings only if you are sure of implications.

    For avoiding config duplication ensure trying to reuse
    [`RedisConnection`][kyotsu.config.databases.models.RedisConnection] first, and specify REDIS_DB only.

    Attributes:
        REDIS_URL: URL to redis server used to store the schedule
        REDIS_DB:
        REDIS_USE_SSL:
        KEY_PREFIX:
        LOCK_KEY:
        LOCK_TIMEOUT:
    """

    REDIS_URL: RedisDsn | None
    REDIS_DB: int | None
    REDIS_USE_SSL: bool = False
    KEY_PREFIX: str = "redbeat"
    LOCK_KEY: str = "redbeat:lock"
    LOCK_TIMEOUT: int = 1500



__all__ = [
    "configure_logger",
]

import logging
from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING

import structlog

from kyotsu.logging.error_codes.processors import error_code_adder
from kyotsu.logging.groups.processors import log_group_adder
from kyotsu.logging.opensearch.processors import OpenSearchLogProcessor
from kyotsu.logging.telemetry.processors import error_telemetry_adder

if TYPE_CHECKING:
    from kyotsu.config.databases import OpenSearchNoCertConnection


def _extract_from_record(
    _: structlog.typing.WrappedLogger,
    __: str,
    event_dict: structlog.typing.EventDict,
) -> structlog.typing.EventDict:
    """
    Processor that will try to extract `thread_name` and `process_name`
    from uvicorn logger, and add them to `event_dict`.
    """
    try:
        record = event_dict["_record"]
    except KeyError:
        return event_dict
    try:
        event_dict["thread_name"] = record.getattr("threadName", event_dict.get("thread_name", "n/a"))
        event_dict["process_name"] = record.getattr("processName", event_dict.get("thread_name", "n/a"))
    except AttributeError:
        return event_dict
    return event_dict


def _apply_to_default_logging(shared_processors: Sequence[structlog.typing.Processor], *, is_dev: bool) -> None:
    """
    Apply structlog logging to default loggers.

    Attributes:
        shared_processors: Previously configured processors chain.
        is_dev: Whether to use [structlog.dev.ConsoleRenderer][] or [structlog.processors.JSONRenderer][] for rendering.
    """
    handler = logging.StreamHandler()

    processors: list[structlog.typing.Processor] = [
        _extract_from_record,
        structlog.stdlib.ProcessorFormatter.remove_processors_meta,
    ]

    if is_dev:
        processors.append(structlog.dev.ConsoleRenderer(colors=True))
    else:
        processors.append(structlog.processors.JSONRenderer())

    formatter = structlog.stdlib.ProcessorFormatter(
        foreign_pre_chain=shared_processors,
        processors=processors,
    )

    handler.setFormatter(formatter)
    root_logger = logging.getLogger()
    root_logger.addHandler(handler)
    root_logger.setLevel(logging.INFO)

    from celery.utils.log import get_task_logger
    celery_logger = get_task_logger("celery.consumer")
    celery_logger.addHandler(handler)
    celery_logger.setLevel(logging.INFO)



def configure_logger(
    *,
    is_dev: bool,
    opensearch_config: "OpenSearchNoCertConnection",
    log_group_index_mapping: Mapping[str, str] | None,
    default_index: str,
    async_only: bool = False,
    cached_logger: bool = False,
) -> None:
    """
    Configure structlog logger.

    This method should be called in that main application file,
    so on app start-up per se.

    Attributes:
        is_dev: True if you're running project locally (doesn't mean `dev` stage).
        opensearch_config: OpenSearch connection config.
        log_group_index_mapping: Mapping of log groups to their respective indices in OpenSearch.
                                 You can keep it None, if you want to fallback on default index only.
        default_index: Default index to use for logs, if log group not found in mapping.
        async_only: Whether to use [structlog.stdlib.AsyncBoundLogger][] logger or not.
        cached_logger: Whether to cache logger or not.
                       (If you are planing to reconfigure logger or use it with multiprocessing keep it `False`)

    Processors:
        Chain of processors used by logger after configuration.

        | Processor | Definition | Description |
        | :-------- | :--------- | :---------- |
        | `timestamper` | [structlog.processors.TimeStamper][] | Add ISO UTC timestamp |
        | `event_to_msg` | [structlog.processors.EventRenamer][] | Rename `event` ro `msg` and then `_event` to `event` |
        | `add_log_level` | [structlog.stdlib.add_log_level][] | Add log level (info, warning, error, etc.) |
        | `add_logger_name` | [structlog.stdlib.add_logger_name][] | Add logger name (passed on initialization) |
        | `set_exc_info` | [structlog.dev.set_exc_info][] | If log_level is `exception` set `exc_info` to `True` |
        | `merge_contextvars` | [structlog.contextvars.merge_contextvars][] | Merge contextvars to `event_dict` (support for bindable loggers). Mainly used for [kyotsu.logging.telemetry.utils][] dcontext managers and decorators |
        | `CallsiteParameter` | [structlog.processors.CallsiteParameterAdder][] | Add `pathname`, `filename`, `func_name`, `thread`, `thread_name`, `process`, `process_name` parameters to `event_dict` |
        | `error_code_adder` | [kyotsu.logging.error_codes.processors.error_code_adder][] | Try to extract error code from [kyotsu.logging.exceptions.BaseKyotsuError][] and add it to `event_dict.error_details` |
        | `log_group_adder` | [kyotsu.logging.groups.processors.log_group_adder][] | Set `group` parameter to `event_dict`. The priority of sources are:  `event_dict` `override_group` parameter; [kyotsu.logging.exceptions.BaseKyotsuError][] `group` parameter; `event_dict` `group` parameter; default group |
        | `error_telemetry_adder` | [kyotsu.logging.telemetry.processors.error_telemetry_adder][] | Try to extract telemetry from [kyotsu.logging.exceptions.BaseKyotsuError][] and add it to `event_dict.error_details` |
        | `ExtraAdder` | [structlog.stdlib.ExtraAdder][] | Unpack `event_dict.extra` to `event_dict` |
        | `dict_tracebacks` | [structlog.processors.dict_tracebacks][] | If `is_dev` is `False` - convert exceptions to suitable for [kyotsu.logging.opensearch.processors.OpenSearchLogProcessor][] and [structlog.processors.JSONPrenderer][] dicts |
        | `OpenSearchLogProcessor` | [kyotsu.logging.opensearch.processors.OpenSearchLogProcessor][] | If `is_dev` is `False` - Send logs to OpenSearch |
        | `ProcessorFormatter.wrap_for_formatter` | [structlog.stdlib.ProcessorFormatter.wrap_for_formatter][] | Wrap logger with `ProcessorFormatter`. Added to support functionality of wrapping classic logger |
        | `_extract_from_record` | [kyotsu.logging.config._extract_from_record][] | Extract `thread_name` and `process_name` from `_record` and add them to `event_dict`. Added to support uvicorn loggers |
        | `ProcessorFormatter.remove_processors_meta` | [structlog.stdlib.ProcessorFormatter.remove_processors_meta][] | Remove `_processors_meta` from `event_dict`. Added to support functionality of wrapping classic logger |
        | `renderer` | [structlog.dev.ConsoleRenderer][] &#124; [structlog.processors.JSONRenderer][] | Use pretty [structlog.dev.ConsoleRenderer][] when `is_dev` is `True`. [structlog.processors.JSONRenderer][] otherwise |
    """
    _log_group_index_mapping: Mapping[str, str] = log_group_index_mapping or {}

    timestamper = structlog.processors.TimeStamper(fmt="ISO", utc=True)
    event_to_msg = structlog.processors.EventRenamer("msg", "_event")

    shared_processors: list[structlog.typing.Processor] = [
        timestamper,  # Add ISO UTC timestamp
        event_to_msg,  # Rename `event` ro `msg` and then `_event` to `event`
        structlog.stdlib.add_log_level,  # Add log level (info, warning, error, etc.)
        structlog.stdlib.add_logger_name,  # Add logger name (passed on initialization)
        structlog.dev.set_exc_info,  # If log_level is `exception` set `exc_info` to `True`
        structlog.contextvars.merge_contextvars,
        structlog.processors.CallsiteParameterAdder({
            structlog.processors.CallsiteParameter.PATHNAME,
            structlog.processors.CallsiteParameter.FILENAME,
            structlog.processors.CallsiteParameter.FUNC_NAME,
            structlog.processors.CallsiteParameter.THREAD,
            structlog.processors.CallsiteParameter.THREAD_NAME,
            structlog.processors.CallsiteParameter.PROCESS,
            structlog.processors.CallsiteParameter.PROCESS_NAME,
        }),
        error_code_adder,
        log_group_adder,
        error_telemetry_adder,
        structlog.stdlib.ExtraAdder(),
    ]

    if not is_dev:
        shared_processors.extend([
            structlog.processors.dict_tracebacks,
            OpenSearchLogProcessor(opensearch_config, _log_group_index_mapping, default_index=default_index),
        ])

    shared_processors.append(structlog.stdlib.ProcessorFormatter.wrap_for_formatter)

    structlog.configure(
        processors=shared_processors,
        wrapper_class=structlog.stdlib.AsyncBoundLogger if async_only else structlog.stdlib.BoundLogger,
        context_class=dict,
        logger_factory=structlog.stdlib.LoggerFactory(),
        cache_logger_on_first_use=cached_logger,
    )

    _apply_to_default_logging(shared_processors, is_dev=is_dev)



__all__ = [
    "TableNameMixin",
    "TimeLoggedMixinDDL",
    "TimeLoggedMixinORM",
    "UUIDMixin",
    "get_onupdate_function",
    "get_table_trigger",
]

import uuid
from datetime import UTC, datetime
from typing import TYPE_CHECKING, Self

import inflection
from sqlalchemy import DateTime, FetchedValue, func, text
from sqlalchemy.orm import Mapped, declared_attr, mapped_column

if TYPE_CHECKING:
    from alembic_utils.pg_function import PGFunction  # type: ignore[import-untyped]
    from alembic_utils.pg_trigger import PGTrigger  # type: ignore[import-untyped]


class TableNameMixin:
    @declared_attr.directive
    @classmethod
    def __tablename__(cls) -> str:
        return inflection.pluralize(inflection.underscore(cls.__name__))


class TimeLoggedMixinORM:
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True),
        default=lambda: datetime.now(UTC),
        nullable=False,
        comment="Timestamp of creation",
    )
    updated_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True),
        default=lambda: datetime.now(UTC),
        onupdate=lambda: datetime.now(UTC),
        nullable=False,
        comment="Timestamp of last update",
    )


class TimeLoggedMixinDDL:
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True),
        server_default=func.now(),
        nullable=False,
        comment="Timestamp of creation",
    )
    updated_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True),
        server_default=func.now(),
        server_onupdate=FetchedValue(),
        nullable=False,
        comment="Timestamp of last update",
    )


def get_onupdate_function(schema: str = "public", column: str = "updated_at") -> "PGFunction":
    from alembic_utils.pg_function import PGFunction

    return PGFunction(
        schema=schema,
        signature=f"update_{column}_column()",
        definition=(
            f"RETURNS TRIGGER AS $$ "
            f"BEGIN "
            f"    NEW.{column} = now(); "
            f"    RETURN NEW; "
            f"END; "
            f"$$ language 'plpgsql';"
        ),
    )


def get_table_trigger(tablename: str, schema: str = "public", column: str = "updated_at") -> "PGTrigger":
    from alembic_utils.pg_trigger import PGTrigger

    return PGTrigger(
        schema=schema,
        signature=f"trg_{tablename}__{column}_auto",
        on_entity=f"{schema}.{tablename}",
        is_constraint=False,
        definition=(
            f"    BEFORE UPDATE "
            f"    ON {tablename} "
            f"    FOR EACH ROW "
            f"EXECUTE PROCEDURE "
            f"    update_{column}_column();"
        ),
    )


class UUIDMixin:
    id: Mapped[uuid.UUID] = mapped_column(
        primary_key=True,
        default=uuid.uuid4,
        server_default=text("gen_random_uuid()"),
        comment="Unique UUID identifier",
    )

    def __str__(self: Self) -> str:
        return f"{self.__class__.__name__} ({self.id})"



__all__ = [
    "log_group_adder",
]

from structlog.processors import _figure_out_exc_info
from structlog.typing import EventDict, WrappedLogger

from kyotsu.logging.groups.constants import LoggingGroup


def _get_override_group_or_none(event_dict: EventDict) -> str | None:
    try:
        return str(event_dict.pop("override_group"))
    except KeyError:
        return None


def _get_exc_group_or_none(event_dict: EventDict) -> str | None:
    exc_info = _figure_out_exc_info(event_dict.get("exc_info"))

    if not exc_info or all(e is None for e in exc_info):
        return None

    return getattr(exc_info[1], "group", None)


def _get_group_or_default(event_dict: EventDict) -> str:
    return event_dict.get("group", LoggingGroup.DEFAULT)


def log_group_adder(_: WrappedLogger, __: str, event_dict: EventDict) -> EventDict:
    event_dict["group"] = (
        _get_override_group_or_none(event_dict)  # trying to get override_group
        or _get_exc_group_or_none(event_dict)  # looking for group in exception
        or _get_group_or_default(event_dict)  # validating group in event dict or returning default group
    )
    return event_dict



__all__ = [
    "AsyncTelemetryClient",
    "TelemetryClient",
]

import typing
from collections.abc import Mapping

from httpx import USE_CLIENT_DEFAULT, AsyncClient, Client, Headers, Response
from httpx._client import UseClientDefault
from httpx._types import (
    AuthTypes,
    CookieTypes,
    HeaderTypes,
    QueryParamTypes,
    RequestContent,
    RequestData,
    RequestExtensions,
    RequestFiles,
    TimeoutTypes,
    URLTypes,
)

from kyotsu.logging.telemetry.contextvars import span_id, trace_id


def _standardize_headers(
    headers: typing.Iterable[tuple[str, str]] | typing.Iterable[tuple[bytes, bytes]],
) -> typing.Iterator[tuple[str, str]]:
    for key, value in headers:
        if isinstance(key, bytes):
            key = key.decode("utf-8")
        if isinstance(value, bytes):
            value = value.decode("utf-8")
        yield key, value


def _update_headers(headers: HeaderTypes | None) -> Headers | dict[str, str]:
    telemetry_headers: dict[str, str] = {
        "X-Trace-ID": trace_id.get() or "",
        "X-Span-ID": span_id.get() or "",
    }
    match headers:
        case False | None:
            # headers are None or empty
            return telemetry_headers
        case Headers():
            for key, value in telemetry_headers.items():
                headers[key] = value
            return headers
        case Mapping():
            return dict(_standardize_headers(headers.items())) | telemetry_headers
        case typing.Sequence():
            return dict(_standardize_headers(headers)) | telemetry_headers


class TelemetryClient(Client):
    def request(
        self,
        method: str,
        url: URLTypes,
        *,
        content: RequestContent | None = None,
        data: RequestData | None = None,
        files: RequestFiles | None = None,
        json: typing.Any | None = None,
        params: QueryParamTypes | None = None,
        headers: HeaderTypes | None = None,
        cookies: CookieTypes | None = None,
        auth: AuthTypes | UseClientDefault | None = USE_CLIENT_DEFAULT,
        follow_redirects: bool | UseClientDefault = USE_CLIENT_DEFAULT,
        timeout: TimeoutTypes | UseClientDefault = USE_CLIENT_DEFAULT,
        extensions: RequestExtensions | None = None,
    ) -> Response:
        return super().request(
            method,
            url,
            content=content,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=_update_headers(headers),
            cookies=cookies,
            auth=auth,
            follow_redirects=follow_redirects,
            timeout=timeout,
            extensions=extensions,
        )


class AsyncTelemetryClient(AsyncClient):
    async def request(
        self,
        method: str,
        url: URLTypes,
        *,
        content: RequestContent | None = None,
        data: RequestData | None = None,
        files: RequestFiles | None = None,
        json: typing.Any | None = None,
        params: QueryParamTypes | None = None,
        headers: HeaderTypes | None = None,
        cookies: CookieTypes | None = None,
        auth: AuthTypes | UseClientDefault | None = USE_CLIENT_DEFAULT,
        follow_redirects: bool | UseClientDefault = USE_CLIENT_DEFAULT,
        timeout: TimeoutTypes | UseClientDefault = USE_CLIENT_DEFAULT,
        extensions: RequestExtensions | None = None,
    ) -> Response:
        return await super().request(
            method,
            url,
            content=content,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=_update_headers(headers),
            cookies=cookies,
            auth=auth,
            follow_redirects=follow_redirects,
            timeout=timeout,
            extensions=extensions,
        )



"""
This module contains types that are used for `kyotsu.config.databases` `mixins` and `models`.

Attributes:
    ConnType: TypeVar that used for generic [`DBConnection`][kyotsu.config.databases.models.DBConnection] models.

Classes:
    AmqpQueryParams: Represents the set of query params used for AMQP connection.
    OpenSearchConnectionDict: Represents the dictionary of connection parameters needed for an OpenSearch connection.
"""

__all__ = [
    "AmqpQueryParams",
    "ConnType",
]

from collections.abc import Sequence
from ssl import SSLContext
from typing import Literal, NotRequired, TypedDict, TypeVar

ConnType = TypeVar("ConnType")
"""Type for `CONN_STR` in classes inherited from generic `DBConnection`."""


class AmqpQueryParams(TypedDict, total=False):
    """
    This `TypedDict` subclass represents the set of query parameters that can be used on an AMQP connection.

    These parameters include detailed AMQP client configurations, such as backpressure detection, retry delay,
    socket timeout and many others.

    This dict should be used as the `QUERY_PARAMS` attribute of an `AmqpConnection` instance.

    See more details at: [pika.connection.URLParameters](https://pika.readthedocs.io/en/stable/modules/parameters.html#pika.connection.URLParameters)
    """

    backpressure_detection: Literal["t"]
    channel_max: int
    connection_attempts: int
    frame_max: int
    heartbeat: int
    locale: str
    ssl_options: dict[str, str]
    retry_delay: int
    socket_timeout: int


class OpenSearchConnectionDict(TypedDict, total=False):
    """
    This `TypedDict` subclass represents the dictionary of connection parameters needed for an OpenSearch connection.

    It is used in the `CONN_DICT` attribute of an `OpenSearchNoCertConnection` instance after being processed in
    `assemble_conn_dict` method.

    This dict establishes details about hosts, authentication, scheme, port, and SSL context, which all will be used to
    configure and establish the OpenSearch connection.
    """

    hosts: Sequence[str]
    http_auth: tuple[str, str] | None
    scheme: Literal["http", "https"]
    port: str

    ssl_context: NotRequired[SSLContext]
    timeout: NotRequired[int]
    max_retries: NotRequired[int]
    retry_on_timeout: NotRequired[bool]
